@@ -131,8 +131,11 @@
 #include "ut.h"
 #include "serialize.h"
 #include "statistics.h"
+#include "status_report.h"
 #include "core_stats.h"
 #include "pvar.h"
+#include "signals.h"
+#include "shutdown.h"
 #include "poll_types.h"
 #include "net/net_tcp.h"
 #include "net/net_udp.h"
@@ -144,38 +147,42 @@
 #include "net/trans.h"
 
 #include "test/unit_tests.h"
+#include "lib/dbg/profiling.h"
 
 #include "ssl_tweaks.h"
 
-/*
- * when enabled ("-T <module>" cmdline param), OpenSIPS will behave as follows:
- *   - enable debug mode
- *   - fork workers normally
- *   - run all currently enabled unit tests
- *     (if module != "core", the modules/<module>/test/ suite is ran,
- *      otherwise the core's ./test/ suite)
- *   - print the unit test summary
- *   - exit with 0 on success, non-zero otherwise
- */
-int testing_framework;
-char *testing_module = "core";
+#include "main_script.h"
+
+/* global vars */
+
+#ifdef VERSION_NODATE
+static char compiled[] =  "" ;
+#else
+#ifdef VERSION_DATE
+static const char compiled[] =  VERSION_DATE ;
+#else
+static const char compiled[] =  __TIME__ " " __DATE__ ;
+#endif
+#endif
+
+static int own_pgid = 0; /* whether or not we have our own pgid (and it's ok
+			    to use kill(0, sig) */
 
 static char* version=OPENSIPS_FULL_VERSION;
 static char* flags=OPENSIPS_COMPILE_FLAGS;
-#ifdef VERSION_NODATE
-char compiled[] =  "" ;
-#else
-#ifdef VERSION_DATE
-const char compiled[] =  VERSION_DATE ;
-#else
-const char compiled[] =  __TIME__ " " __DATE__ ;
-#endif
-#endif
+
+static int user_id = 0;
+static int group_id = 0;
+
+const struct scm_version core_scm_ver = {
+	.type = VERSIONTYPE,
+	.rev = THISREVISION
+};
 
 /**
  * Print compile-time constants
  */
-void print_ct_constants(void)
+static void print_ct_constants(void)
 {
 #ifdef ADAPTIVE_WAIT
 	printf("ADAPTIVE_WAIT_LOOPS=%d, ", ADAPTIVE_WAIT_LOOPS);
@@ -185,584 +192,52 @@ void print_ct_constants(void)
 		MAX_RECV_BUFFER_SIZE, MAX_LISTEN, MAX_URI_SIZE,
 		BUF_SIZE );
 	printf("poll method support: %s.\n", poll_support);
-#ifdef VERSIONTYPE
-	printf("%s revision: %s\n", VERSIONTYPE, THISREVISION);
-#endif
+	printf("%s revision: %s\n", core_scm_ver.type, core_scm_ver.rev);
 }
 
-/* global vars */
-
-int own_pgid = 0; /* whether or not we have our own pgid (and it's ok
-					 to use kill(0, sig) */
-char* cfg_file = 0;
-char *preproc = NULL;
-unsigned int maxbuffer = MAX_RECV_BUFFER_SIZE; /* maximum buffer size we do
-												  not want to exceed during the
-												  auto-probing procedure; may
-												  be re-configured */
-/* number of UDP workers processing requests */
-int udp_workers_no = UDP_WORKERS_NO;
-/* the global UDP auto scaling profile */
-char *udp_auto_scaling_profile = NULL;
-/* if the auto-scaling engine is enabled or not - this is autodetected */
-int auto_scaling_enabled = 0;
-/* auto-scaling sampling and checking time cycle is 1 sec by default */
-int auto_scaling_cycle = 1;
-/*!< by default choose the best method */
-enum poll_types io_poll_method=0;
-/* last signal received */
-int sig_flag = 0;
-
-/* activate debug mode */
-int debug_mode = 0;
-/* do not become daemon, stay attached to the console */
-int no_daemon_mode = 0;
-/* assertion statements in script. disabled by default */
-int enable_asserts = 0;
-/* abort process on failed assertion. disabled by default */
-int abort_on_assert = 0;
-/* start by logging to stderr */
-int log_stderr = 1;
-/* log facility (see syslog(3)) */
-int log_facility = LOG_DAEMON;
-/* the id to be printed in syslog */
-char *log_name = 0;
-int config_check = 0;
-/* check if reply first via host==us */
-int check_via =  0;
-/* debugging level for memory stats */
-int memlog = L_DBG + 11;
-int memdump = L_DBG + 10;
-/* debugging in case msg processing takes. too long disabled by default */
-int execmsgthreshold = 0;
-/* debugging in case dns takes too long. disabled by default */
-int execdnsthreshold = 0;
-/* debugging in case tcp stuff take too long. disabled by default */
-int tcpthreshold = 0;
-/* should replies include extensive warnings? by default yes,
-   good for trouble-shooting
-*/
-int sip_warning = 0;
-/* should localy-generated messages include server's signature? */
-int server_signature=1;
-/* Server header to be used when proxy generates a reply as UAS.
-   Default is to use SERVER_HDR CRLF (assigned later).
-*/
-str server_header = {SERVER_HDR,sizeof(SERVER_HDR)-1};
-/* User-Agent header to be used when proxy generates request as UAC.
-   Default is to use USER_AGENT CRLF (assigned later).
-*/
-str user_agent_header = {USER_AGENT,sizeof(USER_AGENT)-1};
-/* should opensips try to locate outbound interface on multihomed
- * host? by default not -- too expensive
- */
-int mhomed=0;
-/* use dns and/or rdns or to see if we need to add
-   a ;received=x.x.x.x to via: */
-int received_dns = 0;
-char* working_dir = 0;
-char* chroot_dir = 0;
-char* user=0;
-char* group=0;
-int user_id = 0;
-int group_id = 0;
-
-/* more config stuff */
-int disable_core_dump=0; /* by default enabled */
-int open_files_limit=-1; /* don't touch it by default */
-
-#ifdef USE_MCAST
-int mcast_loopback = 0;
-int mcast_ttl = -1; /* if -1, don't touch it, use the default (usually 1) */
-#endif /* USE_MCAST */
-
-int tos = IPTOS_LOWDELAY; // lgtm [cpp/short-global-name]
-
-struct socket_info* bind_address=0; /* pointer to the crt. proc.
-									 listening address*/
-
-
-/* if aliases should be automatically discovered and added
- * during fixing listening sockets */
-int auto_aliases=0;
-
-/* if the stateless forwarding support in core should be
- * disabled or not */
-int sl_fwd_disabled=-1;
-
-/* process number - 0 is the main process */
-int process_no = 0;
-
-/* cfg parsing */
-int cfg_errors=0;
-
-/* start-up time */
-time_t startup_time = 0;
-
-/* shared memory (in MB) */
-unsigned long shm_mem_size=SHM_MEM_SIZE * 1024 * 1024;
-unsigned int shm_hash_split_percentage = DEFAULT_SHM_HASH_SPLIT_PERCENTAGE;
-unsigned int shm_secondary_hash_size = DEFAULT_SHM_SECONDARY_HASH_SIZE;
-
-/* packaged memory (in MB) */
-unsigned long pkg_mem_size=PKG_MEM_SIZE * 1024 * 1024;
-
-
-/* export command-line to anywhere else */
-int my_argc;
-char **my_argv;
-
-int is_main = 1; /* flag = is this the  "main" process? */
-
-/* flag = is this an initial, pre-daemon process ? */
-int is_pre_daemon = 1;
-
-char* pid_file = 0; /* filename as asked by user */
-char* pgid_file = 0;
-
-
-/**
- * Clean up on exit. This should be called before exiting.
- * \param show_status set to one to display the mem status
- */
-void cleanup(int show_status)
+static int
+init_suid(void)
 {
-	LM_INFO("cleanup\n");
-
-	/*clean-up*/
-
-	/* hack: force-unlock the shared memory lock(s) in case
-	   		 some process crashed and let it locked; this will
-	   		 allow an almost gracious shutdown */
-	if (0
-#if defined F_MALLOC || defined Q_MALLOC
-		|| mem_lock
-#endif
-#ifdef HP_MALLOC
-		|| mem_locks
-#endif
-	) {
-#if defined HP_MALLOC && (defined F_MALLOC || defined Q_MALLOC)
-		if (mem_allocator_shm == MM_HP_MALLOC ||
-		        mem_allocator_shm == MM_HP_MALLOC_DBG) {
-			int i;
-
-			for (i = 0; i < HP_HASH_SIZE; i++)
-				lock_release(&mem_locks[i]);
-		} else {
-			shm_unlock();
-		}
-#elif defined HP_MALLOC
-		int i;
-
-		for (i = 0; i < HP_HASH_SIZE; i++)
-			lock_release(&mem_locks[i]);
-#else
-		shm_unlock();
-#endif
-	}
-
-	handle_ql_shutdown();
-	destroy_modules();
-	udp_destroy();
-	tcp_destroy();
-	destroy_timer();
-	destroy_stats_collector();
-	destroy_script_cb();
-	pv_free_extra_list();
-	tr_free_extra_list();
-	destroy_argv_list();
-	destroy_black_lists();
-	free_route_lists(sroutes); // this is just for testing purposes
-#ifdef PKG_MALLOC
-	if (show_status){
-		LM_GEN1(memdump, "Memory status (pkg):\n");
-		pkg_status();
-	}
-#endif
-	cleanup_log_level();
-
-	if (pt && (0
-#if defined F_MALLOC || defined Q_MALLOC
-		|| mem_lock
-#endif
-#ifdef HP_MALLOC
-		|| mem_locks
-#endif
-		))
-		shm_free(pt);
-	pt=0;
-	if (show_status){
-			LM_GEN1(memdump, "Memory status (shm):\n");
-			shm_status();
-	}
-
-	/* zero all shmem alloc vars that we still use */
-	shm_mem_destroy();
-	if (pid_file) unlink(pid_file);
-	if (pgid_file) unlink(pgid_file);
+	/* all processes should have access to all the sockets (for sending)
+	 * so we open all first*/
+	return do_suid(user_id, group_id);
 }
 
-
-/**
- * Send a signal to all child processes
- * \param signum signal for killing the children
- */
-static void kill_all_children(int signum)
+static int
+init_shm_mem_dbg(void)
 {
-	int r;
-
-	if (!pt)
-		return;
-
-	for (r = 1; r < counted_max_processes; r++) {
-		if (pt[r].pid == -1 || (pt[r].flags & OSS_PROC_DOING_DUMP))
-			continue;
-
-		/* as the PIDs are filled in by child processes, a 0 PID means
-		 * an un-initalized procees; killing an uninitialized proc is
-		 * very dangerous, so better wait for it to finish its init
-		 * sequence by blocking until the pid is populated */
-		while (pt[r].pid == 0)
-			usleep(1000);
-
-		kill(pt[r].pid, signum);
-	}
-}
-
-
-/**
- * SIGALRM "timeout" handler during the attendant's final cleanup,
- * try to leave a core for future diagnostics.
- */
-static void sig_alarm_abort(int signo)
-{
-	/* LOG is not signal safe, but who cares, we are abort-ing anyway :-) */
-	LM_CRIT("BUG - shutdown timeout triggered, dying...\n");
-	abort();
-}
-
-
-/* RPC function send by main process to all worker processes supporting
- * IPC in order to force a gracefull termination
- */
-static void rpc_process_terminate(int sender_id, void *code)
-{
-	#ifdef PKG_MALLOC
-	LM_GEN1(memdump, "Memory status (pkg):\n");
-	pkg_status();
-	#endif
-
-	/* simply terminate the process */
-	LM_DBG("Process %d exiting with code %d...\n",
-		process_no, (int)(long)code);
-
-	exit( (int)(long)code );
-}
-
-
-/* Implements full shutdown sequence (terminate processes and cleanup)
- * To be called ONLY from MAIN process, not from workers !!!
- */
-static void shutdown_opensips( int status )
-{
-	pid_t  proc;
-	int i, n, p;
-	int chld_status;
-
-	set_osips_state( STATE_TERMINATING );
-
-	/* terminate all processes */
-
-	/* first we try to terminate the processes via the IPC channel */
-	for( i=1,n=0 ; i<counted_max_processes; i++) {
-		/* Depending on the processes status, its PID may be:
-		 *   -1 - process not forked yet
-		 *    0 - process forked but not fully configured by core
-		 *   >0 - process fully running
-		 */
-		if (pt[i].pid!=-1) {
-			/* use IPC (if avaiable) for a graceful termination */
-			if ( IPC_FD_WRITE(i)>0 ) {
-				LM_DBG("Asking process %d [%s] to terminate\n", i, pt[i].desc);
-				if (ipc_send_rpc( i, rpc_process_terminate, (void*)0)<0) {
-					LM_ERR("failed to trigger RPC termination for "
-						"process %d\n", i );
-				}
-			} else {
-				while (pt[i].pid==0) usleep(1000);
-				kill(pt[i].pid, SIGTERM);
-			}
-			n++;
-		}
-	}
-
-	/* now wait for the processes to finish */
-	i = GRACEFUL_SHUTDOWN_TIMEOUT * 100;
-	while( i && n ) {
-		proc = waitpid( -1, &chld_status, WNOHANG);
-		if (proc<=0) {
-			/* no process exited so far, do a small sleep before retry */
-			usleep(10000);
-			i--;
-		} else {
-			if ( (p=get_process_ID_by_PID(proc)) == -1 ) {
-				LM_DBG("unknown child process %d ended. Ignoring\n",proc);
-			} else {
-				LM_INFO("process %d(%d) [%s] terminated, "
-					"still waiting for %d more\n", p, proc, pt[p].desc, n-1);
-				/* mark the child process as terminated / not running */
-				pt[p].pid = -1;
-				status |= chld_status;
-				n--;
-			}
-		}
-	}
-
-	if (i==0 && n!=0) {
-		LM_DBG("force termination for all processes\n");
-		kill_all_children(SIGKILL);
-	}
-
-	/* Only one process is running now. Clean up and return overall status */
-	signal(SIGALRM, sig_alarm_abort);
-	alarm(SHUTDOWN_TIMEOUT - i / 100);
-	cleanup(1);
-	alarm(0);
-	signal(SIGALRM, SIG_IGN);
-
-	dprint("Thank you for running " NAME "\n");
-	exit( status );
-}
-
-
-/**
- * Signal handler for the server.
- */
-void handle_sigs(void)
-{
-	pid_t  chld;
-	int    chld_status,overall_status=0;
-	int    i;
-	int    do_exit;
-
-	switch(sig_flag){
-		case 0: break; /* do nothing*/
-		case SIGPIPE:
-				/* SIGPIPE might be rarely received on use of
-				   exec module; simply ignore it
-				 */
-		case SIGUSR1:
-		case SIGUSR2:
-		case SIGHUP:
-				/* ignoring it*/
-				break;
-		case SIGINT:
-		case SIGTERM:
-			/* we end the program in all these cases */
-			if (sig_flag==SIGINT)
-				LM_DBG("SIGINT received, program terminates\n");
-			else
-				LM_DBG("SIGTERM received, program terminates\n");
-
-			shutdown_opensips( 0/*status*/ );
-			break;
-
-		case SIGCHLD:
-			do_exit = 0;
-			while ((chld=waitpid( -1, &chld_status, WNOHANG ))>0) {
-				/* is it a process we know about? */
-				if ( (i=get_process_ID_by_PID( chld )) == -1 ) {
-					LM_DBG("unknown child process %d ended. Ignoring\n",chld);
-					continue;
-				}
-				if (pt[i].flags & OSS_PROC_SELFEXIT) {
-					LM_NOTICE("process %d/%d did selfexit with "
-						"status %d\n", i, chld,  WTERMSIG(chld_status));
-					reset_process_slot(i);
-					continue;
-				}
-				do_exit = 1;
-				/* process the signal */
-				overall_status |= chld_status;
-				LM_DBG("OpenSIPS exit status = %d\n",overall_status);
-
-				if (WIFEXITED(chld_status))
-					LM_INFO("child process %d exited normally,"
-							" status=%d\n", chld,
-							WEXITSTATUS(chld_status));
-				else if (WIFSIGNALED(chld_status)) {
-					LM_INFO("child process %d exited by a signal"
-							" %d\n", chld, WTERMSIG(chld_status));
-#ifdef WCOREDUMP
-					LM_INFO("core was %sgenerated\n",
-							 WCOREDUMP(chld_status) ?  "" : "not " );
-#endif
-				}else if (WIFSTOPPED(chld_status))
-					LM_INFO("child process %d stopped by a"
-								" signal %d\n", chld,
-								 WSTOPSIG(chld_status));
-
-				/* mark the child process as terminated / not running */
-				pt[i].pid = -1;
-			}
-			if (!do_exit)
-				break;
-			LM_INFO("terminating due to SIGCHLD\n");
-			/* exit */
-			shutdown_opensips( overall_status );
-			break;
-
-		default:
-			LM_CRIT("unhandled signal %d\n", sig_flag);
-	}
-	sig_flag=0;
-}
-
-/* the initial SIGSEGV handler, provided by the OS */
-static struct sigaction sa_sys_segv;
-static char sa_sys_is_valid;
-
-static inline int restore_segv_handler(void)
-{
-	LM_DBG("restoring SIGSEGV handler...\n");
-
-	if (!sa_sys_is_valid)
-		return 1;
-
-	if (sigaction(SIGSEGV, &sa_sys_segv, NULL) < 0) {
-		LM_ERR("failed to restore system SIGSEGV handler\n");
-		return -1;
-	}
-
-	LM_DBG("successfully restored system SIGSEGV handler\n");
-
+	if (shm_memlog_size)
+		shm_mem_enable_dbg();
 	return 0;
 }
 
-/**
- * Exit regulary on a specific signal.
- * This is good for profiling which only works if exited regularly
- * and not by default signal handlers
- * \param signo The signal that should be handled
- */
-static void sig_usr(int signo)
-{
-	int status;
-	pid_t pid;
-	UNUSED(pid);
-
-	if (is_main){
-		if (signo == SIGSEGV) {
-			LM_CRIT("segfault in attendant (starter) process!\n");
-			if (restore_segv_handler() != 0)
-				exit(-1);
-			return;
-		}
-
-		if (sig_flag == 0)
-			sig_flag = signo;
-		else /*  previous sig. not processed yet, ignoring? */
-			return;
-	}else{
-		/* process the important signals */
-		switch(signo){
-			case SIGPIPE:
-			case SIGINT:
-			case SIGUSR1:
-			case SIGUSR2:
-			case SIGHUP:
-					/* ignored*/
-					break;
-			case SIGTERM:
-					/* ignore any SIGTERM if not in shutdown sequance (this 
-					 * is marked by the attendent process) */
-					if (get_osips_state()!=STATE_TERMINATING)
-						return;
-					/* if some shutdown already in progress, ignore this one */
-					if (sig_flag==0) sig_flag=signo;
-					else return;
-					/* do the termination */
-					LM_INFO("signal %d received\n", signo);
-					/* print memory stats for non-main too */
-					#ifdef PKG_MALLOC
-					LM_GEN1(memdump, "Memory status (pkg):\n");
-					pkg_status();
-					#endif
-					exit(0);
-					break;
-			case SIGCHLD:
-					while ( (pid = waitpid(-1, &status, WNOHANG))>0 );
-					break;
-			case SIGSEGV:
-					/* looks like we ate some spicy SIP */
-					LM_CRIT("segfault in process pid: %d, id: %d\n",
-					        pt[process_no].pid, process_no);
-					pt[process_no].flags |= OSS_PROC_DOING_DUMP;
-					if (restore_segv_handler() != 0)
-						exit(-1);
-					pkg_status();
-		}
-	}
-}
-
-
-
-/**
- * Install the signal handlers.
- * \return 0 on success, -1 on error
- */
-int install_sigs(void)
-{
-	struct sigaction act;
-
-	memset(&act, 0, sizeof act);
-
-	act.sa_handler = sig_usr;
-	if (sigaction(SIGSEGV, &act, &sa_sys_segv) < 0) {
-		LM_INFO("failed to install custom SIGSEGV handler -- corefiles must "
-		        "now be written within %d sec to avoid truncation!\n",
-		        GRACEFUL_SHUTDOWN_TIMEOUT);
-	} else {
-		LM_DBG("override SIGSEGV handler: success\n");
-		sa_sys_is_valid = 1;
-	}
-
-	if (signal(SIGINT, sig_usr) == SIG_ERR ) {
-		LM_ERR("no SIGINT signal handler can be installed\n");
-		goto error;
-	}
-	/* if we debug and write to a pipe, we want to exit nicely too */
-	if (signal(SIGPIPE, sig_usr) == SIG_ERR ) {
-		LM_ERR("no SIGINT signal handler can be installed\n");
-		goto error;
-	}
-
-	if (signal(SIGUSR1, sig_usr)  == SIG_ERR ) {
-		LM_ERR("no SIGUSR1 signal handler can be installed\n");
-		goto error;
-	}
-	if (signal(SIGCHLD , sig_usr)  == SIG_ERR ) {
-		LM_ERR("no SIGCHLD signal handler can be installed\n");
-		goto error;
-	}
-	if (signal(SIGTERM , sig_usr)  == SIG_ERR ) {
-		LM_ERR("no SIGTERM signal handler can be installed\n");
-		goto error;
-	}
-	if (signal(SIGHUP , sig_usr)  == SIG_ERR ) {
-		LM_ERR("no SIGHUP signal handler can be installed\n");
-		goto error;
-	}
-	if (signal(SIGUSR2 , sig_usr)  == SIG_ERR ) {
-		LM_ERR("no SIGUSR2 signal handler can be installed\n");
-		goto error;
-	}
-	return 0;
-error:
-	return -1;
-}
-
+static const struct main_script main_script[] = {
+	FN_HNDLR(init_reactor_size, <, 0, "internal reactor"),
+	FN_HNDLR(init_timer, <, 0, "timer"),
+	FN_HNDLR(init_shm_mem_dbg, <, 0, "shm_mem_enable_dbg"),
+	FN_HNDLR(init_ipc, <, 0, "IPC support"),
+	FN_HNDLR(init_serialization, !=, 0, "serialization"),
+	FN_HNDLR(init_mi_core, <, 0, "MI core"),
+	FN_HNDLR(evi_register_core, !=, 0, "register core events"),
+	FN_HNDLR(init_black_lists, !=, 0, "black list engine"),
+	FN_HNDLR(resolv_blacklist_init, !=, 0, "resolver's blacklist"),
+	FN_HNDLR(init_dset, !=, 0, "SIP forking logic"),
+	FN_HNDLR(init_db_support, !=, 0, "SQL database support"),
+	FN_HNDLR(init_cdb_support, !=, 0, "CacheDB support"),
+	FN_HNDLR(init_modules, !=, 0, "modules"),
+	FN_HNDLR(init_xlog, <, 0, "xlog"),
+	FN_HNDLR(register_route_timers, <, 0, "route_timers"),
+	FN_HNDLR(init_pvar_support, !=, 0, "pseudo-variable support"),
+	FN_HNDLR(init_multi_proc_support, !=, 0, "multi processes support"),
+	FN_HNDLR(init_extra_avps, !=, 0, "avps"),
+	FN_HNDLR(fix_rls, !=, 0, "routing lists"),
+	FN_HNDLR(init_log_level, !=, 0, "logging levels"),
+	FN_HNDLR(init_log_event_cons, <, 0, "log event consumer"),
+	FN_HNDLR(trans_init_all_listeners, <, 0, "all SIP listeners"),
+	FN_HNDLR(init_script_reload, <, 0, "cfg reload ctx"),
+	FN_HNDLR(init_suid, ==, -1, "do_suid"),
+	FN_HNDLR(run_post_fork_handlers, <, 0, "post-fork handlers"),
+};
 
 /**
  * Main loop, forks the children, bind to addresses,
@@ -775,8 +250,14 @@ static int main_loop(void)
 	int* startup_done = NULL;
 	utime_t last_check = 0;
 	int rc;
+	static struct internal_fork_handler profiling_handler = {
+		.desc = "_ProfilerStart_child()",
+		.post_fork.in_child = _ProfilerStart_child,
+		.post_fork.in_parent = _ProfilerStart_parent,
+	};
 
 	chd_rank=0;
+	register_fork_handler(&profiling_handler);
 
 	if (start_module_procs()!=0) {
 		LM_ERR("failed to fork module processes\n");
@@ -831,12 +312,13 @@ static int main_loop(void)
 		shm_free(startup_done);
 	}
 
-	set_osips_state( STATE_RUNNING );
+	sr_set_core_status( STATE_RUNNING, MI_SSTR("running"));
+	sr_add_core_report( MI_SSTR("initialization completed, ready now") );
 
 	/* main process left */
 	is_main=1;
 	set_proc_attrs("attendant");
-	pt[process_no].flags |= OSS_PROC_NO_IPC|OSS_PROC_NO_LOAD;
+	pt[process_no].flags |= OSS_PROC_IS_RUNNING|OSS_PROC_NO_IPC|OSS_PROC_NO_LOAD;
 
 	if (testing_framework) {
 		if (init_child(1) < 0) {
@@ -864,6 +346,11 @@ static int main_loop(void)
 		last_check = get_uticks();
 	}
 
+	if (set_log_event_cons_cfg_state() < 0) {
+		LM_ERR("Failed to set the config state for event consumer\n");
+		goto error;
+	}
+
 	for(;;){
 			handle_sigs();
 			if (auto_scaling_enabled) {
@@ -900,9 +387,7 @@ error:
  */
 int main(int argc, char** argv)
 {
-	/* configure by default logging to syslog */
-	int cfg_log_stderr = 1;
-	int c,r;
+	int c, n;
 	char *tmp;
 	int tmp_len;
 	int port;
@@ -912,6 +397,7 @@ int main(int argc, char** argv)
 	int ret;
 	unsigned int seed;
 	int rfd;
+	int procs_no;
 
 	/*init*/
 	ret=-1;
@@ -920,7 +406,7 @@ int main(int argc, char** argv)
 	/* process pkg mem size from command line */
 	opterr=0;
 
-	options="f:cCm:M:b:l:n:N:rRvdDFEVhw:t:u:g:p:P:G:W:o:a:k:s:"
+	options="A:f:cCm:M:b:l:n:N:rRvdDFEVhw:t:u:g:p:P:G:W:o:a:k:s:"
 #ifdef UNIT_TESTS
 	"T:"
 #endif
@@ -943,7 +429,7 @@ int main(int argc, char** argv)
 					}
 					break;
 			case 'd':
-					*log_level = *log_level + 1;
+					*log_level = debug_mode ? L_DBG : (*log_level)+1;
 					break;
 			case 'u':
 					user=optarg;
@@ -1000,7 +486,7 @@ int main(int argc, char** argv)
 	if (init_pkg_mallocs()==-1)
 		goto error00;
 
-	if ( (sroutes=new_sroutes_holder())==NULL )
+	if ( (sroutes=new_sroutes_holder(0))==NULL )
 		goto error00;
 
 	/* we want to be sure that from now on, all the floating numbers are 
@@ -1022,7 +508,6 @@ int main(int argc, char** argv)
 					if (config_check==3)
 						break;
 					config_check |= 1;
-					cfg_log_stderr=1; /* force stderr logging */
 					break;
 			case 'm':
 			case 'M':
@@ -1077,8 +562,9 @@ int main(int argc, char** argv)
 					no_daemon_mode=1;
 					break;
 			case 'E':
-					cfg_log_stderr=1;
-					break;
+					LM_ERR("-E option deprecated since 3.4, set \"stderror_enabled=yes\" "
+					    "at the script level instead\n");
+					goto error00;
 			case 'N':
 					tcp_workers_no=strtol(optarg, &tmp, 10);
 					if ((tmp==0) ||(*tmp)){
@@ -1132,6 +618,10 @@ int main(int argc, char** argv)
 					if (add_arg_var(optarg) < 0)
 						LM_ERR("cannot add option %s\n", optarg);
 					break;
+			case 'A':
+					default_global_address->s = optarg;
+					default_global_address->len = strlen(optarg);
+					break;
 #ifdef UNIT_TESTS
 			case 'T':
 					LM_INFO("running in testing framework mode, for '%s'\n", optarg);
@@ -1158,8 +648,6 @@ int main(int argc, char** argv)
 		}
 	}
 
-	log_stderr = cfg_log_stderr;
-
 	/* seed the prng, try to use /dev/urandom if possible */
 	/* no debugging information is logged, because the standard
 	   log level prior the config file parsing is L_NOTICE */
@@ -1207,7 +695,13 @@ try_again:
 		goto error;
 	}
 
-	set_osips_state( STATE_STARTING );
+	if (init_status_report() < 0) {
+		LM_ERR("failed to initialize status-report support\n");
+		goto error;
+	}
+
+	sr_set_core_status( STATE_INITIALIZING, MI_SSTR("initializing"));
+	sr_add_core_report( MI_SSTR("initializing") );
 
 	if ((!testing_framework || strcmp(testing_module, "core"))
 	        && parse_opensips_cfg(cfg_file, preproc, NULL) < 0) {
@@ -1215,8 +709,19 @@ try_again:
 		goto error00;
 	}
 
+	if (shm_memlog_size && init_dbg_shm_mallocs()==-1)
+		goto error;
+
 	/* shm statistics, module stat groups, memory warming */
-	init_shm_post_yyparse();
+	if (init_shm_post_yyparse() != 0) {
+		LM_ERR("failed to init SHM memory\n");
+		return ret;
+	}
+
+	if (init_log_cons_shm_table() < 0) {
+		LM_ERR("Failed to initialize shm log consumers table\n");
+		goto error;
+	}
 
 	if (config_check>1 && check_rls()!=0) {
 		LM_ERR("bad function call in config file\n");
@@ -1300,9 +805,15 @@ try_again:
 			LM_NOTICE("disabling daemon mode (found enabled)\n");
 			no_daemon_mode = 1;
 		}
-		if (log_stderr==0) {
+		if (stderr_enabled==0) {
 			LM_NOTICE("enabling logging to standard error (found disabled)\n");
-			log_stderr = 1;
+			stderr_enabled = 1;
+			set_log_consumer_mute_state(&str_init(STDERR_CONSUMER_NAME), 0);
+		}
+		if (syslog_enabled) {
+			LM_NOTICE("disabling logging to syslog (found enabled)\n");
+			syslog_enabled = 0;
+			set_log_consumer_mute_state(&str_init(SYSLOG_CONSUMER_NAME), 1);
 		}
 		if (*log_level < L_DBG && (!testing_framework ||
 		                           !strcmp(testing_module, "core"))) {
@@ -1322,10 +833,11 @@ try_again:
 			}
 		}
 		if (tcp_count_processes(NULL)!=0) {
-			if (tcp_workers_no!=2) {
-				LM_NOTICE("setting TCP children to 2 (found %d)\n",
-					tcp_workers_no);
-				tcp_workers_no = 2;
+			procs_no = (tcp_workers_max_no>=2) ? 2 : tcp_workers_max_no;
+			if (tcp_workers_no != procs_no) {
+				LM_NOTICE("setting TCP children to %d (found %d)\n",
+					procs_no, tcp_workers_no);
+				tcp_workers_no = procs_no;
 			}
 		}
 
@@ -1363,131 +875,15 @@ try_again:
 	LM_NOTICE("using system memory for private process memory\n");
 #endif
 
-	/* init async reactor */
-	if (init_reactor_size()<0) {
-		LM_CRIT("failed to init internal reactor, exiting...\n");
-		goto error;
+	for (n = 0; n < howmany(main_script, main_script[0]); n++) {
+		int result = main_script[n].hndlr();
+		pred_cmp_f pred_cmp = cmps_ops[main_script[n].pval];
+		if (pred_cmp(result, main_script[n].pred)) {
+			LM_ERR("failed to initialize %s!\n", main_script[n].desc);
+			goto error;
+		}
 	}
 
-	/* init timer */
-	if (init_timer()<0){
-		LM_CRIT("could not initialize timer, exiting...\n");
-		goto error;
-	}
-
-	/* init IPC */
-	if (init_ipc()<0){
-		LM_CRIT("could not initialize IPC support, exiting...\n");
-		goto error;
-	}
-
-	/* init serial forking engine */
-	if (init_serialization()!=0) {
-		LM_ERR("failed to initialize serialization\n");
-		goto error;
-	}
-	/* Init MI */
-	if (init_mi_core()<0) {
-		LM_ERR("failed to initialize MI core\n");
-		goto error;
-	}
-
-	/* Register core events */
-	if (evi_register_core() != 0) {
-		LM_ERR("failed register core events\n");
-		goto error;
-	}
-
-	/* init black list engine */
-	if (init_black_lists()!=0) {
-		LM_CRIT("failed to init blacklists\n");
-		goto error;
-	}
-	/* init resolver's blacklist */
-	if (resolv_blacklist_init()!=0) {
-		LM_CRIT("failed to create DNS blacklist\n");
-		goto error;
-	}
-
-	if (init_dset() != 0) {
-		LM_ERR("failed to initialize SIP forking logic!\n");
-		goto error;
-	}
-
-	/* init SQL DB support */
-	if (init_db_support() != 0) {
-		LM_ERR("failed to initialize SQL database support\n");
-		goto error;
-	}
-
-	/* init CacheDB support */
-	if (init_cdb_support() != 0) {
-		LM_ERR("failed to initialize CacheDB support\n");
-		goto error;
-	}
-
-	/* init modules */
-	if (init_modules() != 0) {
-		LM_ERR("error while initializing modules\n");
-		goto error;
-	}
-
-	/* init xlog */
-	if (init_xlog() < 0) {
-		LM_ERR("error while initializing xlog!\n");
-		goto error;
-	}
-
-	/* register route timers */
-	if(register_route_timers() < 0) {
-		LM_ERR("Failed to register timer\n");
-		goto error;
-	}
-
-	/* init pseudo-variable support */
-	if (init_pvar_support() != 0) {
-		LM_ERR("failed to init pvar support\n");
-		goto error;
-	}
-
-	/* init multi processes support */
-	if (init_multi_proc_support()!=0) {
-		LM_ERR("failed to init multi-proc support\n");
-		goto error;
-	}
-
-	/* init avps */
-	if (init_extra_avps() != 0) {
-		LM_ERR("error while initializing avps\n");
-		goto error;
-	}
-
-	/* fix routing lists */
-	if ( (r=fix_rls())!=0){
-		LM_ERR("failed to fix configuration with err code %d\n", r);
-		goto error;
-	}
-
-	if (init_log_level() != 0) {
-		LM_ERR("failed to init logging levels\n");
-		goto error;
-	}
-
-	if (trans_init_all_listeners()<0) {
-		LM_ERR("failed to init all SIP listeners, aborting\n");
-		goto error;
-	}
-
-	if (init_script_reload()<0) {
-		LM_ERR("failed to init cfg reload ctx, aborting\n");
-		goto error;
-	}
-
-	/* all processes should have access to all the sockets (for sending)
-	 * so we open all first*/
-	if (do_suid(user_id, group_id)==-1)
-		goto error;
-
 	ret = main_loop();
 
 error:
@@ -1497,5 +893,6 @@ error:
 	cleanup(0);
 error00:
 	LM_NOTICE("Exiting....\n");
+	_ProfilerStop();
 	return ret;
 }
