@@ -123,8 +123,7 @@ static inline char *decode_mail_url(char *p, char *p_end, char *url,
 														unsigned char *nr_attr)
 {
 	static char buf[ MAX_EMAIL_HDR_SIZE ];
-	char c;
-	char foo;
+	int c, foo;
 	unsigned short hdr_len;
 	unsigned short *len;
 	int max_len;
@@ -157,7 +156,7 @@ static inline char *decode_mail_url(char *p, char *p_end, char *url,
 					"character in mail url [%.*s]\n", 3, url);
 				goto error;
 			}
-			c = c<<4 | foo;
+			c = (char)(c<<4 | foo);
 			url += 3;
 		} else {
 			/* normal character - just copy it without changing */
