@@ -111,12 +111,16 @@
 #include "t_funcs.h"
 #include "t_reply.h"
 #include "t_cancel.h"
+#include "t_stats.h"
 
 
 static struct timer_table *timertable=0;
 static unsigned int timer_sets = 0;
 static struct timer detached_timer; /* just to have a value to compare with*/
 
+/* currently processed branch */
+extern int _tm_branch_index;
+
 #define DETACHED_LIST (&detached_timer)
 
 #define is_in_timer_list2(_tl) ( (_tl)->timer_list &&  \
@@ -247,6 +251,7 @@ static void fake_reply(struct cell *t, int branch, int code )
 	branch_bm_t cancel_bitmap = 0;
 	enum rps reply_status;
 
+	_tm_branch_index = branch;
 	if ( is_local(t) ) {
 		reply_status=local_reply( t, FAKED_REPLY, branch,
 					  code, &cancel_bitmap );
@@ -257,6 +262,7 @@ static void fake_reply(struct cell *t, int branch, int code )
 		reply_status=relay_reply( t, FAKED_REPLY, branch, code,
 			&cancel_bitmap );
 	}
+	_tm_branch_index = 0;
 	/* again, a final negative reply on a branch will never lead to a
 	 * situation to cancel other existing branches, so the
 	 * cancel_bitmap should be empty here (we use it as a dummy holder), so
@@ -303,12 +309,20 @@ inline static void retransmission_handler( struct timer_link *retr_tl )
 			}
 
 			set_t(T_UNDEFINED);
+			switch(r_buf->retr_list) {
+			case RT_T1_TO_1: update_stat( tm_retran_req_T11, 1); break;
+			case RT_T1_TO_2: update_stat( tm_retran_req_T12, 1); break;
+			case RT_T1_TO_3: update_stat( tm_retran_req_T13, 1); break;
+			case RT_T2: update_stat( tm_retran_req_T2, 1);break;
+			default:;
+			}
 	} else {
 			LM_DBG("retransmission_handler : reply resending "
 				"(t=%p, %.9s ... )\n", r_buf->my_T, r_buf->buffer.s);
 			set_t(r_buf->my_T);
 			t_retransmit_reply(r_buf->my_T);
 			set_t(T_UNDEFINED);
+			update_stat( tm_retran_rpl_T2, 1);
 	}
 
 	id = r_buf->retr_list;
@@ -386,7 +400,7 @@ inline static void final_response_handler( struct timer_link *fr_tl )
 		}
 	}
 	memset( my_ctx, 0, context_size(CONTEXT_GLOBAL) );
-	current_processing_ctx = my_ctx;
+	set_global_context(my_ctx);
 	/* set the T context too */
 	set_t( t );
 
@@ -398,6 +412,12 @@ inline static void final_response_handler( struct timer_link *fr_tl )
 		cancel_uacs(t, cancel_bitmap );
 		set_cancel_extra_hdrs( NULL, 0);
 	}
+
+	switch(r_buf->retr_list) {
+	case FR_TIMER_LIST: update_stat( tm_timeout_fr, 1); break;
+	case FR_INV_TIMER_LIST: update_stat( tm_timeout_fr_inv, 1); break;
+	default:;
+	}
 	/* lock reply processing to determine how to proceed reliably */
 	LOCK_REPLIES( t );
 	LM_DBG("Cancel sent out, sending 408 (%p)\n", t);
@@ -409,7 +429,7 @@ inline static void final_response_handler( struct timer_link *fr_tl )
 	else
 		context_destroy(CONTEXT_GLOBAL, my_ctx);
 	/* switch back to the old context */
-	current_processing_ctx = old_ctx;
+	set_global_context(old_ctx);
 	/* reset the T context */
 	init_t();
 
@@ -595,8 +615,9 @@ struct timer_table *tm_init_timers( unsigned int sets )
 			init_timer_list( set, i );
 
 		/* exclusion timer */
-		if ((timertable[set].ex_lock = lock_init_rw()) == NULL) {
-			LM_CRIT("failed to init timer RW lock\n");
+		if ((timertable[set].ex_lock = lock_alloc()) == NULL ||
+		lock_init(timertable[set].ex_lock) == NULL) {
+			LM_CRIT("failed to init timer lock\n");
 			goto error0;
 		}
 
@@ -626,7 +647,7 @@ void free_timer_table(void)
 		for ( i=0 ; i<timer_sets*NR_OF_TIMER_LISTS ; i++ )
 			release_timerlist_lock( &timertable->timers[i] );
 		for ( i=0 ; i<timer_sets ; i++ )
-			lock_destroy_rw( timertable[i].ex_lock );
+			lock_destroy( timertable[i].ex_lock );
 		shm_free(timertable);
 	}
 }
@@ -1071,9 +1092,12 @@ static void unlink_timers( struct cell *t )
 void timer_routine(unsigned int ticks , void *set)
 {
 	struct timer_link *tl, *tmp_tl;
+	struct timespec begin;
 	int                id;
 
-	lock_start_write( timertable[(long)set].ex_lock );
+	clock_gettime(CLOCK_REALTIME, &begin);
+
+	lock_get( timertable[(long)set].ex_lock );
 
 	for( id=0 ; id<RT_T1_TO_1 ; id++ )
 	{
@@ -1095,7 +1119,11 @@ void timer_routine(unsigned int ticks , void *set)
 				break;
 		}
 	}
-	lock_stop_write( timertable[(long)set].ex_lock );
+	lock_release( timertable[(long)set].ex_lock );
+
+	clock_check_diff((double)TM_TIMER_ITV_S*1e9 * TM_TIMER_LOAD_WARN,
+	    "now at %d%%+ capacity, inuse_transactions: %lu", (int)(TM_TIMER_LOAD_WARN*100),
+	    (unsigned long)get_stat_val(tm_trans_inuse));
 }
 
 
@@ -1103,9 +1131,12 @@ void timer_routine(unsigned int ticks , void *set)
 void utimer_routine(utime_t uticks , void *set)
 {
 	struct timer_link *tl, *tmp_tl;
+	struct timespec begin;
 	int                id;
 
-	lock_start_write( timertable[(long)set].ex_lock );
+	clock_gettime(CLOCK_REALTIME, &begin);
+
+	lock_get( timertable[(long)set].ex_lock );
 
 	for( id=RT_T1_TO_1 ; id<NR_OF_TIMER_LISTS ; id++ )
 	{
@@ -1123,6 +1154,10 @@ void utimer_routine(utime_t uticks , void *set)
 				break;
 		}
 	}
-	lock_stop_write( timertable[(long)set].ex_lock );
+	lock_release( timertable[(long)set].ex_lock );
+
+	clock_check_diff((double)TM_UTIMER_ITV_US*1000 * TM_TIMER_LOAD_WARN,
+	    "now at %d%%+ capacity, inuse_transactions: %lu", (int)(TM_TIMER_LOAD_WARN*100),
+	    (unsigned long)get_stat_val(tm_trans_inuse));
 }
 
