@@ -41,6 +41,7 @@
 #define MSG_PARSER_H
 
 #include <strings.h>
+#include <sys/time.h>
 
 #include "../str.h"
 #include "../lump_struct.h"
@@ -87,7 +88,7 @@ enum request_method {
 };
 
 #define FL_FORCE_RPORT       (1<<0)  /* force rport (top via) */
-#define FL_FORCE_ACTIVE      (1<<1)  /* force active SDP */
+#define FL_REPLY_TO_VIA      (1<<1)  /* force replying to VIA ip:port */
 #define FL_FORCE_LOCAL_RPORT (1<<2)  /* force local rport (local via) */
 #define FL_SDP_IP_AFS        (1<<3)  /* SDP IP rewritten */
 #define FL_SDP_PORT_AFS      (1<<4)  /* SDP port rewritten */
@@ -123,6 +124,13 @@ enum request_method {
                                       * route */
 #define FL_TM_REPLICATED	 (1<<19) /* message received due to a tm replication */
 #define FL_BODY_NO_SDP       (1<<20) /* message does not have an SDP body */
+#define FL_IS_LOCAL          (1<<21) /* the message is a locally generated
+                                      * one, not received */
+#define FL_HAS_ROUTE_LUMP    (1<<22) /* the message had Route headers added
+                                      * as lumps */
+#define FL_USE_SIPTRACE_B2B  (1<<23) /* used by tracer to check if the b2b
+                                      * tracing was enabled */
+#define FL_ACK_WITH_BODY     (1<<24) /* ACK message has SDP body */
 
 /* define the # of unknown URI parameters to parse */
 #define URI_MAX_U_PARAMS 10
@@ -201,6 +209,9 @@ struct sip_uri {
 	str pn_prid_val;
 	str pn_param_val;
 	str pn_purr_val;
+	/* XXX - in the future when adding params as special links
+	 * in the list above, make sure to also update compare_uris() function
+	 * to explicitly compare these here */
 
 	/* unknown params */
 	str u_name[URI_MAX_U_PARAMS]; /* Unknown param names */
@@ -271,6 +282,9 @@ struct sip_msg {
 	struct hdr_field* min_expires;
 	struct hdr_field* feature_caps;
 	struct hdr_field* replaces;
+	struct hdr_field* security_client;
+	struct hdr_field* security_server;
+	struct hdr_field* security_verify;
 
 	struct sip_msg_body *body;
 
@@ -293,7 +307,7 @@ struct sip_msg {
 	unsigned int ruri_bflags; /* per-branch flags for RURI*/
 
 	/* force sending on this socket */
-	struct socket_info* force_send_socket;
+	const struct socket_info* force_send_socket;
 
 	/* path vector to generate Route hdrs */
 	str path_vec;
@@ -332,6 +346,12 @@ struct sip_msg {
 	str set_global_address;
 	str set_global_port;
 
+	/* used to store a particular time of the message - note that the time is
+	 * not stored when the message was received, but only the first time
+	 * someone gets interested in it - this way we have a consistent timestamp
+	 * of the message, without being affected by lazy callbacks */
+	struct timeval time;
+
 	struct msg_callback *msg_cb;
 };
 
@@ -346,9 +366,16 @@ extern int via_cnt;
 
 int parse_msg(char* buf, unsigned int len, struct sip_msg* msg);
 
-int parse_headers(struct sip_msg* msg, hdr_flags_t flags, int next);
+#define parse_headers(msg, flags,next) 	parse_headers_aux(msg,flags,next, 1)
 
-char* get_hdr_field(char* buf, char* end, struct hdr_field* hdr);
+int parse_headers_aux(struct sip_msg* msg, hdr_flags_t flags, int next, int sip_well_known_parse);
+
+#define get_hdr_field(buf,end,hdr)	get_hdr_field_aux(buf,end,hdr,1)
+
+char* get_hdr_field_aux(char* buf, char* end, struct hdr_field* hdr, int sip_well_known_parse);
+
+/* add DEL lumps for all headers matching the given @hdr */
+int delete_headers(struct sip_msg *msg, struct hdr_field *hdr);
 
 void free_sip_msg(struct sip_msg* msg);
 
@@ -425,12 +452,14 @@ inline static int get_body(struct sip_msg *msg, str *body)
 	body->len = msg->buf + msg->len - body->s;
 
 	/* double check the len against content-length hdr
-	   (if present, it must be already parsed) */
+	   (if present, it must be already parsed);
+	   NOTE that the CT hdr may be missing if using UDP, so 
+	   we do not consider its missing an err case */
 	if (msg->content_length) {
 		ct_len = get_content_length( msg );
 		if (ct_len<body->len)
 			body->len = ct_len;
-	} else {
+	} else if (msg->rcv.proto!=PROTO_UDP) {
 		/* no ct -> no body */
 		body->s = NULL;
 		body->len = 0;
@@ -481,6 +510,23 @@ inline static struct hdr_field *get_header_by_name( struct sip_msg *msg,
 }
 
 
+#define get_next_header_by_static_name(_hdr, _name) \
+		get_next_header_by_name(_hdr, _name, sizeof(_name)-1)
+inline static struct hdr_field *get_next_header_by_name(
+						struct hdr_field *first, char *s, unsigned int len)
+{
+	struct hdr_field *hdr;
+
+	for( hdr=first->next ; hdr ; hdr=hdr->next ) {
+		if(len==hdr->name.len && strncasecmp(hdr->name.s,s,len)==0)
+			return hdr;
+	}
+	return NULL;
+}
+
+
+
+
 /*
  * Make a private copy of the string and assign it to new_uri (new RURI)
  */
@@ -555,4 +601,16 @@ void clear_path_vector(struct sip_msg* msg);
  */
 int extract_ftc_hdrs( char *buf, int len, str *from, str *to, str *cseq,str *callid);
 
+inline static struct timeval *get_msg_time(struct sip_msg *msg)
+{
+	static struct timeval static_time;
+	if (!msg || msg == (struct sip_msg *)-1) {
+		gettimeofday(&static_time, NULL);
+		return &static_time;
+	}
+	if (msg->time.tv_sec == 0 && msg->time.tv_usec == 0)
+		gettimeofday(&msg->time, NULL);
+	return &msg->time;
+}
+
 #endif
